"""
ScurvySin
---------
Simple tool that installs a package getting as much
dependencies as possible using `conda` and `pip`
in the remaining items from the dependency tree.
"""

import json
import os
import subprocess
import sys
from pkg_resources import Requirement, Environment
from typing import Dict, List


__version__ = "0.1.2"


class CondaFlags:
    """Class representing optional flags for conda requested by user"""

    def __init__(self, args):
        """
        transform args given to the main program to anaconda flags
        """
        self.dry_run = args.dry_run

    def expfl(self) -> list:
        if self.dry_run:
            return ['--dry-run']
        else:
            return []


class PipFlags:
    """Class representing optional flags for pip requested by user"""

    def __init__(self, args):
        """
        transform args given to the main program to pip flags
        """
        self.dry_run = args.dry_run

    def expfl(self) -> list:
        # pip does not have a --dry-run option, but download is more or less
        # a dry run counterpart to install.
        if self.dry_run:
            return ['download']
        else:
            return ['install']


def already_satisfied(req: str) -> bool:
    requirement = Requirement(req)
    environment = Environment()
    return any(dist in requirement for dist in environment[requirement.name])


def available_in_conda(req: str) -> bool:
    r = subprocess.run(["conda", "install", "-d", req],
                       stderr=subprocess.PIPE,
                       stdout=subprocess.PIPE)
    return not(r.returncode)


def install_using_conda(req: str, flags: CondaFlags) -> None:
    subprocess.call(["conda", "install", "-S", "-y"] + flags.expfl() + [req])


def install_using_pip(req: str, flags: PipFlags) -> None:
    subprocess.call(["pip"] +  flags.expfl() + ["--no-deps", req])


def get_pip_requirements(req: str) -> Dict[str, str]:
    abspath = os.path.abspath(os.path.dirname(__file__))
    dep_script = os.path.join(abspath, "get_pip_deps.py")
    r = subprocess.run([sys.executable, dep_script, req],
                       stdout=subprocess.PIPE)
    try:
        data = json.loads(r.stdout)
        if "error" in data:
            print(f"Error: {data['error']}")
            exit(1)
        else:
            return data["requirements"]
    except json.decoder.JSONDecodeError as err:
        print("Invalid JSON")
        exit(1)


def parse_requirements_file(path: str) -> List[str]:
    reqs = []
    with open(path, "r") as infile:
        for line in infile:
            line = line.strip()
            if line.startswith("-"):
                raise NotImplementedError("Pip flags in requirements files are not understood.")
            elif line.startswith("#"):
                continue
            else:
                reqs.append(line)
    return list(sorted(reqs))


def try_install(req: str, opts: dict, coflags: CondaFlags, pipflags: PipFlags) -> None:
    if opts.pop("requirement", False):
        print(f"Reading requirements file {req}")
        requirements = parse_requirements_file(req)
        print(f"Dependencies from {req}: {requirements}.")
        for requirement in requirements:
            try_install(requirement, opts, coflags, pipflags)
        return
    if already_satisfied(req):
        print(f"Condition {req} already satistfied.")
        return
    print(f"Checking {req} in conda...")
    if available_in_conda(req):
        print(f"Package {req} found in conda.")
        if opts["show_only"]:
            print(f"Would install {req} using conda.")
        else:
            print(f"Installing {req} using conda.")
            install_using_conda(req, coflags)
    else:
        print(f"Checking dependencies for {req} using pip...")
        requirements = get_pip_requirements(req)
        print(f"Dependencies for {req}: {list(requirements.values())}.")
        for requirement in requirements.values():
            try_install(requirement, opts, coflags, pipflags)
        if opts["show_only"]:
            print(f"Would install {req} using pip.")
        else:
            print(f"Installing {req} using pip.")
            install_using_pip(req, pipflags)



#!/usr/bin/env python3
"""Script to extract a list of first-level dependencies using pip.

Warning: This file is not meant to be imported as a module.

Examples:
    $ python get_pip_deps.py pandas==0.24

        {
            "requirements": {
                "pytz": "pytz>=2011k",
                "numpy": "numpy>=1.12.0",
                "python-dateutil": "python-dateutil>=2.5.0"
            }
        }

    $ python get_pip_deps.py n0nex1st3nt

        {
            "error": "Not found."
        }
"""
import json
import os
import sys
import tempfile
import warnings
from io import StringIO
from typing import List

# Sin: Importing from protected modules
from pip._internal import main
from pip._internal.resolve import Resolver, ensure_dir
from pip._internal.req.req_install import InstallRequirement


def get_dependencies(r: str) -> List[InstallRequirement]:
    """Get direct dependencies for a requirement string.

    :param r: Requirement - it can be a package name or a name
       with version specification.

    It runs the equivalent of `pip download` command
    with a monkey-patched resolver that stops at the
    first level of dependency tree.

    This hack works, but keep this in mind:
        1) Monkey patching is a bad idea in general.
        2) Relying on anything starting with a "_" is a bad idea.
        3) Pip documentation explicitly discourages attempts to use it as library.

    The resolver is simplified a bit - we removed the parts
    that are not used.
    
    Examples:
        >>> get_dependencies("pandas==0.24")
        [<InstallRequirement object: numpy>=1.12.0 (from pandas==0.24) editable=False>,
         <InstallRequirement object: python-dateutil>=2.5.0 (from pandas==0.24) editable=False>,
         <InstallRequirement object: pytz>=2011k (from pandas==0.24) editable=False>]
    """
    old_stdout = sys.stdout
    old_stderr = sys.stderr
    sys.stdout = StringIO()
    sys.stderr = StringIO()

    try:
        found_reqs = []  # type: List[InstallRequirement]

        # Sin: Monkey-patching
        def new_resolve(self, requirement_set):
            # make the wheelhouse
            if self.preparer.wheel_download_dir:
                ensure_dir(self.preparer.wheel_download_dir)

            # If any top-level requirement has a hash specified, enter
            # hash-checking mode, which requires hashes from all.
            root_reqs = (
                requirement_set.unnamed_requirements +
                list(requirement_set.requirements.values())
            )
            self.require_hashes = (
                requirement_set.require_hashes or
                any(req.has_hash_options for req in root_reqs)
            )

            for req in root_reqs:
                found_reqs.extend(self._resolve_one(requirement_set, req))

        Resolver.resolve = new_resolve

        try:
            with tempfile.TemporaryDirectory() as tmpdir:
                os.chdir(tmpdir)
                main(["download", r, "-qqqqq"])
        except Exception:
            pass

        if "Could not find" in sys.stderr.getvalue():
            raise RuntimeError("Not found.")

    finally:
        sys.stdout = old_stdout
        sys.stderr = old_stderr

    found_reqs.sort(key=lambda l: l.name)
    return found_reqs


def run():
    """Get the dependencies and write them to stdout in JSON format.

    It takes the first argument from the command line.
    """
    req = sys.argv[1]
    try:
        deps = get_dependencies(req)
        data = {
            "requirements": {
                dep.name: str(dep.req) for dep in deps
            }
        }
    except RuntimeError as exc:
        data = {
            "error": str(exc)
        }
    print(json.dumps(data, indent=2))


if __name__ == "__main__":
    run()
else:
    warnings.warn("`get_pip_deps` is not meant to be imported as module.")



from setuptools import setup, find_packages

from scurvysin import __doc__, __version__


options = dict(
    name='scurvysin',
    version=__version__,
    packages=find_packages(),
    license='MIT',
    description='Pip & conda wrapper / warper.',
    long_description=__doc__.strip(),
    author='Jan Pipek, Pavel Cahyna',
    author_email='jan.pipek@gmail.com',
    url='https://github.com/pythomec/scurvysin',
    python_requires="~=3.6",
    entry_points = {
        'console_scripts' : [
            "scurvy=scurvysin.cli:main"
        ]
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
    ]
)
setup(**options)




#!/usr/bin/env python3
import sys
import argparse

from scurvysin import try_install, CondaFlags, PipFlags


def extopts(args: argparse.Namespace):
    return {
        "show_only": args.show_only,
        "requirement": args.requirement,
    }


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--dry-run', '-d', action='store_true',
                        help="don't let conda/pip perform installation")
    parser.add_argument('--show-only', action='store_true',
                        help="only show what would be done")
    parser.add_argument('--requirement', '-r', action='store_true',
                        help="use requirements file")
    parser.add_argument('req', help="distribution to install")
    args = parser.parse_args()

    try_install(args.req, extopts(args), CondaFlags(args), PipFlags(args))


if __name__ == "__main__":
    main()
